{{
ParallelDevice.spin - Parallel Propeller Device communications driver
Copyright (c) 2009 by John Steven Denson (jazzed)
See MIT license terms below. 
}}

con
  CMD_RDBUF     = 0             ' host cmd -1
  CMD_WRBUF     = 1
  CMD_RUNRPC    = 2
  CMD_REGRPC    = 3
  CMD_NOP       = 4
  CMD_NOTIFY    = 5
  
obj
  def: "ParallelDefs"
  cs : "CommandStruct"

var
  long cog
    
pub start(id)
  stop
  cs.setPropid(id)
  cs.setTablePtr(@cmdtable)
  cs.setCommand(CMD_NOP)
  cog := cognew(@entry, cs.getCommandPtr) + 1
  cs.waitCommand
  return cog

pub stop
  if cog
    cogstop(cog~ -1)

pub notify | back
{{
  Notify lets host know we have a message to send.
}}
  cs.setCommand(CMD_NOP)  ' any value in command will cause asynchronous interrupt
  cs.waitCommand
  waitcnt(clkfreq/((?back)//10+10)+cnt) ' restrict the number of interrupts per second

    
dat
'----------------------------------------------------------------------
entry          org      0
  long         0 [8] ' nops will be replaced by PASM debugger stub
  jmp          #init

' pointers in this order
'
cmdptr         long 0   ' COG interface command address
tabptr         long 0   ' COG interface command table pointer address
dstptr         long 0   ' COG interface destination address
adrptr         long 0   ' COG interface address address
lenptr         long 0   ' COG interface buffer length address
bufptr         long 0   ' COG interface buffer data address
pidptr         long 0   ' COG interface Propeller ID address

cmd            long 0   ' one ofcommand table index commands
tab            long 0   ' command table commands
dst            long 0   ' the Secondary Propeller ID and COG target
adr            long 0   ' command address
len            long 0   ' length of buffer to send / receive
bufp           long 0   ' buffer start
pid            long 0   ' Propeller Chip ID

tp             long 0   ' temporary pointer
t0             long 0   ' temporary variable
t1             long 0   ' temporary variable
t2             long 0   ' temporary variable

ack            long 1 << def#ACKPIN     ' always input  for host/device
atn            long 1 << def#ATNPIN     ' always output for host/device
int            long 1 << def#INTPIN     ' always output for device
dbus           long $ff                 ' output at command phase. in/out on data.
dbusack        long $ff | (1<<def#ACKPIN)
dbusackstp     long $ff | (1<<def#ACKPIN) | (1 << def#STPBIT)
stp            long 1 << def#STPBIT
ackstp         long (1 << def#ACKPIN) | (1 << def#STPBIT)
zero           long  0
leds           long $1 << 15

'----------------------------------------------------------------------
init
  mov          tp,      par
  mov          cmdptr,  tp              ' cmd changes always
  add          tp,      #4
  mov          tabptr,  tp              ' only need to get this once
  add          tp,      #4
  mov          dstptr,  tp              ' dst changes always
  add          tp,      #4
  mov          adrptr,  tp              ' adr changes always
  add          tp,      #4
  mov          lenptr,  tp              ' len changes always
  add          tp,      #4
  mov          bufptr,  tp              ' buf changes always
  add          tp,      #4
  mov          pidptr,   tp             ' our propeller chip id

  rdlong       pid,     pidptr          ' do this at startup only?
  and          pid,     #def#PIDMSK

  or           dira,    leds

  
'----------------------------------------------------------------------
' Host asserts ATT for any command to get device attention.
' First ATT byte contains device id
' Second byte contains cog id and request.
' Next contains addr low, then next contains addr high.
' Last header byte contains length of data.
' Rest of bytes contain data.
'
' BYTE0  DEVID[8]
' BYTE1  COGID[4]CMDREQ[4]
' BYTE2  ADDRLOW[8]
' BYTE3  ADDRHIGH[8]
' BYTE4  LENGTH[8]
' BYTEN  DATA0..N ... for RPC commands, this data is a packet
'
' For header, ACK low means device is ready.
' For transactions, ACK high signals ready for data transfer.
'
' Host Protocol:
' 1 BYTE0 on bus
' 2 Assert ATN low
' 3 Wait for ACK low
' 4 Disassert ATN to high
' 5 Send BYTE1,2,3,4
' 6 Wait for ACK low

' Device Protocol:
' 1 Wait for ATN low
' 2 Check BYTE0 for PropID match
' 3 Assert ACK low
' 4 Prepare to get BYTE1,2,3,4
' 5 Wait for ATN high
' 6 Get BYTE1,2,3,4
' 7 Assert ACK low
' 8 in command assert ACK high
    
'----------------------------------------------------------------------
nextcmd        ' do next cmd
  andn         dira,    dbusackstp      ' disassert dbus, ack, and stp bit
  wrlong       zero,    cmdptr          ' let startup and notify know we're ready
nextloop
  ' 1 Wait for ATN low
  test         atn,     ina wz            
  if_z  jmp    #handler                 ' handle ATT
  rdlong       cmd,     cmdptr wz
  if_nz jmp    #async                   ' notify host of event
  jmp          #nextloop                ' nothing to do
handler   
  mov          dst,     ina             ' 2 get BYTE0
  and          dst,     #def#PIDMSK     ' check only chip id

  ' 2 is message for me ?
  cmp          dst,     pid wc,wz       
  if_ne jmp    #nextcmd                 ' 2 message not for me
   
  ' got atn and PropID match

  ' 3 assert ack low 
  andn         outa,    ack             ' 3 message is for me ... ACK low
  or           dira,    ack             ' 3 ACK low to let host know we got it
  
  ' 4 prepare for bytes
  ' 5 wait atn high
  waitpne      ack,     ack             ' sync-up only
  mov          bufp,    #0              ' zero values
  mov          t0,      #0              ' zero values
  waitpeq      atn,     atn             ' 5 wait for atn high

  ' 6 get BYTE1,2,3,4 read header
  or           outa,    ack             ' OK give me the command                get it
  movs         cmd,     ina             ' BYTE1 - command[0:3] and cogid[4:7]
  andn         outa,    ack             ' low
  or           outa,    ack             ' ready for next byte                   get it
  movs         bufp,    ina             ' BYTE2 - low address
  andn         outa,    ack             ' low
  or           outa,    ack             ' ready for next byte                   get it
  movs         t0,      ina             ' BYTE3 - high address
  andn         outa,    ack             ' low
  or           outa,    ack             ' ready for next byte                   get it
  movs         len,     ina             ' BYTE4 - data length
  andn         outa,    ack             ' low
  and          cmd,     #7              ' make sure only bits[0:2] are set
  shl          cmd,     #1              ' command table is word wide
  and          t0,      dbus            ' only 8 bits please
  shl          t0,      #8              ' compile address
  and          bufp,    dbus            ' only 8 bits please
  or           bufp,    t0              ' tp is the target address
  rdlong       t1,      tabptr          ' get command table
  add          t1,      cmd             ' add offset for command
  and          len,     dbus            ' only 8 bits please
  rdword       cmd,     t1              ' get command address

  ' 7 Assert ack low
  andn         outa,    ack             ' ack low
  jmp          cmd                      ' jump to command, returns to nextcmd
  
'----------------------------------------------------------------------
_idle          ' do nothing command used only at startup 
  jmp          #nextcmd

'----------------------------------------------------------------------
_rdbufd        ' read from buffer / write to host
  rdbyte       t0,      t0              ' kill time ... other side has to catch up
  rdbyte       t0,      t0              ' kill time ... other side has to catch up
  rdbyte       t0,      t0              ' kill time ... other side has to catch up
  or           dira,    dbus            ' dbus to output
  andn         dira,    ack             ' transfering data
  rdbyte       t0,      t0              ' kill time ... other side has to catch up
  mov          tp,      bufp            ' set the base address
  add          tp,      len             ' add the length
  rdbyte       t0,      t0              ' sync and kill time ... other side has to catch up
  or           outa,    leds            ' debug only
  andn         outa,    stp             ' set stp low
  or           dira,    stp             ' signal start write ... bit 9 is pulled high externally
  nop                                   ' leave low for detector
  xor          outa,    leds            ' debug only
  andn         dira,    stp             ' high while transfering data
  rdbyte       outa,    tp              ' first byte on the bus
  sub          tp,      #1              ' next byte
'{
:write
  rdbyte       outa,    tp              ' get byte from djnz controlled pointer
  cmp          bufp,    tp wz           ' test for end condition
  if_ne  djnz  tp,      #:write         ' if ptr > addr, decrement pointer and repeat
'}
{                                                  
:write
  xor          outa,    leds            ' debug only
  cmp          bufp,    tp wz           ' test for end condition
  if_ne djnz   tp,      #:write         ' if ptr > addr, decrement pointer and repeat
'}
  andn         outa,    #$1ff           ' clear bits
  or           dira,    stp             ' signal finished with bits[0:8] == 0
  xor          outa,    leds            ' debug only
  nop                                   ' make sure device gets stop message before releasing it
  rdbyte       t0,      t0              ' make sure device gets stop message before releasing it
  jmp          #nextcmd

'----------------------------------------------------------------------
_wrbufd        ' read from host / write to buffer
  mov          tp,      bufp            ' set the base address
  add          tp,      len             ' add the length
  mov          t1,      #0              ' make sure all bits are clear
  or           outa,    leds            ' debug only
:wait
  test         stp,     ina wz          ' wait for stop to go low
  if_nz jmp    #:wait
  'xor          outa,    leds            ' debug only
  nop
  nop
  nop          ' why would device need 3 NOP before read where host only needs 2?
  'rdlong       t0,      t0              ' sync and delay
'{  
:read
  movs         t1,      ina wz          ' looking for 0 on all bits and P[0:8]
  wrbyte       t1,      tp              ' write to ptr = address + length
  if_nz djnz   tp,      #:read          ' if t0 <> 0, continue with ptr-1; else stop 
'}
{                                                  
:read
  xor          outa,    leds            ' debug only
  wrbyte       t0,      tp
  djnz         len,      #:read         ' if t0 <> 0, continue with ptr-1; else stop 
'}
  andn         outa,    leds            ' debug only
  jmp          #nextcmd

'----------------------------------------------------------------------
_runrpcd       ' execute a previously arranged remote procedure function call
  jmp          #nextcmd

'----------------------------------------------------------------------
_regrpcd       ' register a remote procedure function call
  jmp          #nextcmd

'----------------------------------------------------------------------
' When ATN is high, wait for our PID to arrive on the bus ...
' when the host gets an intterupt, it polls all possible sources.
' After the host puts our pid is on the bus, it starts to read us.
' If we have message for the host, we drive our pid on the bus and assert ACK.
' Else we drive bus to 0 (host pid) to indicate no message. 
' 
_poll           ' tell host about a pending message when asked. clear interrupt.

'  mov          t0,      ina             ' get byte
'  and          t0,      pidmask         ' only pid bits
'  cmp          pid,     t0 wz           ' compare our pid ...
'  if_nz jmp    #nextcmd                 ' if not our pid, exit

  andn         dira,    int             ' disassert int
  rdbyte       t0,      dstptr wz       ' message to send in dst ?
  andn         outa,    dbus            ' more P8 reset trouble
  if_nz or     outa,    pid             ' set propid for message queue
  or           dira,    dbus            ' read from hub, write to host
  or           outa,    ack             ' say data good
  nop                                   ' delay before return
  nop                                   ' delay before return                                                                 
  jmp          #nextcmd                 ' all done

'----------------------------------------------------------------------
' Send an interrupt ... interrupts are cleared by Host sending notify message 
'
' Use separate int line for now ...
'    can we use end of a Host->Device transaction to do this with the ACK signal ?
'
async          ' set asynchronous interrupt from device to host
  andn         outa,    int             ' int low
  or           dira,    int             ' assert int ... _notify disasserts when Host polls
  jmp          #nextcmd                 ' all done


'----------------------------------------------------------------------
dat     ' debug command table
  long         ' align to long
  cmdtable     word _rdbufd     ' 0 CMD_RDBUF    - read from buffer to target
               word _wrbufd     ' 1 CMD_WRBUF    - write to buffer from target
               word _runrpcd    ' 2 CMD_RUN_RPC  - execute a remote function
               word _regrpcd    ' 3 CMD_REGISTER - register a remote function
               word _poll       ' 4 CMD_POLL     - respond to host poll
               word _idle       ' 5 CMD_NOP      - used only at startup
               word _idle       ' 6 CMD_NOP      - used only at startup
               word _idle       ' 7 CMD_NOP      - used only at startup
              
{{
 TERMS OF USE: MIT License                                                           

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
}}