{{
ParallelHost.spin - Parallel Propeller Host communications driver
Copyright (c) 2009 by John Steven Denson (jazzed)
See MIT license terms below. 
}}

obj
  def: "ParallelDefs"
  cs : "CommandStruct"

obj
  bd    : "BMAdebugger"         ' BMA debugger routines
  ot    : "BMAopcodes"          ' required for mnemonic instructions (optional)
  bu    : "BMAutility"          ' BMA debugger utilitiy routines
  s     : "FullDuplexSingleton" ' required for single stepping with main/stepDump
  
var
  long cog

con
  QMSK  = $ff                ' up to 256 notification events in the message queue

var ' asyncronous reverse path queue is circular and should have 2**N entries
  byte  queue[QMSK+1]        ' asynchronous handler message queue
    
pub start(last)
{{
  start the driver using device startup structure
}}
  stop
  cs.setTablePtr(@cmdtable)
  cs.setQueuePtr(@queue)
  cs.setLastID(last)
  cs.setCommand(def#CMD_NOP)
  cog := cognew(@entry, cs.getCommandPtr) + 1
  'cog := bu.debug(@entry, cs.getCommandPtr) + 1
  cs.waitCommand
  return cog

pub stop
{{
  stop this driver
}}
  if cog
    cogstop(cog~ -1)

pub getPasm
{{
  get PASM entry point for debug
}}
  return @entry
  
pub rdbuffer(bptr,size,dest,addr) '(parmp)
{{
  read data buffer from target
}}
  cs.setDataFields(bptr,size,dest,addr)
  repeat while cs.getCommand
  cs.setCommand(def#CMD_RDBUF)
  cs.waitCommand

  
pub wrbuffer(bptr,size,dest,addr) '(parmp)
{{
  write data buffer to target
}}
  cs.setDataFields(bptr,size,dest,addr)
  repeat while cs.getCommand
  cs.setCommand(def#CMD_WRBUF)
  cs.waitCommand

pub flushMessage
  repeat while hasMessage
     getMessage
     
pub hasMessage
{{
  return non-zero if queue head <> tail
}}
  return long[cs.getQtailPtr] <> long[cs.getQheadPtr]

pub getMessage | hdp, tlp
  hdp := cs.getQheadPtr
  tlp := cs.getQtailPtr
  if long[tlp] <> long[hdp]
    long[tlp]++
    long[tlp] := long[tlp] & QMSK
    
  return byte[@queue][long[tlp]]


dat
'----------------------------------------------------------------------
entry          org      0
  long         0 [8] ' nops will be replaced by PASM debugger stub
  jmp          #init

' pointers in this order
'
lastid         long 0   ' last propeller device on the bus

cmdptr         long 0   ' COG interface command address
tabptr         long 0   ' COG interface command table pointer address
dstptr         long 0   ' COG interface destination address
adrptr         long 0   ' COG interface address address
lenptr         long 0   ' COG interface buffer length address
bufptr         long 0   ' COG interface buffer data address
pidptr         long 0   ' data structure contains pidptr before queue stuff
lidptr         long 0   ' data structure contains lastidptr before queue stuff
qptr           long 0   ' queue address
hptr           long 0   ' queue head address
tptr           long 0   ' queue tail address

cmd            long 0   ' one ofcommand table index commands
tab            long 0   ' command table commands
dst            long 0   ' the Secondary Propeller ID and COG target
adr            long 0   ' command address
len            long 0   ' length of buffer to send / receive
bufp           long 0   ' buffer start
intsrc         long 0   ' interrupt source for asynchronous events
head           long 0   ' queue head
tail           long 0   ' queue tail

pidmask        long def#PIDMSK

tp             long 0   ' temporary pointer
t0             long 0   ' temporary variable
t1             long 0   ' temporary variable
t2             long 0   ' temporary variable

ack            long 1 << def#ACKPIN             ' always input  for host/device
atn            long 1 << def#ATNPIN             ' always output for host/device
int            long 1 << def#INTPIN             ' always input for host
dbus           long $ff                         ' output at command phase. in/out on data.
dbusatn        long $ff | (1 << def#ATNPIN)     ' output at command phase. in/out on data.
dbusatnstp     long $ff | (1 << def#ATNPIN) | (1 << def#STPBIT) ' adds STPMSK to bus release
zero           long 0
stp            long 1 << def#STPBIT             ' stop mask for 5MB/s data transfers
dbusstp        long $ff | (1 << def#STPBIT)     ' release dbus and stop bit

leds           long $7 << 12

'----------------------------------------------------------------------
init           'jmp      #init
  mov          tp,      par
  mov          cmdptr,  tp              ' cmd changes always
  add          tp,      #4
  mov          tabptr,  tp              ' only need to get this once
  add          tp,      #4
  mov          dstptr,  tp              ' dst changes always
  add          tp,      #4
  mov          adrptr,  tp              ' adr changes always
  add          tp,      #4
  mov          lenptr,  tp              ' len changes always
  add          tp,      #4
  mov          bufptr,  tp              ' buf changes always
  add          tp,      #4              
  mov          pidptr,  tp              ' Propeller Id - can skip
  add          tp,      #4
  mov          lidptr,  tp              ' Last Propeller Id
  add          tp,      #4
  mov          hptr,    tp              ' head pointer
  add          tp,      #4              
  mov          tptr,    tp              ' tail pointer - can skip ?
  add          tp,      #4
  rdlong       qptr,    tp              ' queue pointer
  rdlong       lastid,  lidptr          ' get lastid
  or           dira,    leds

'----------------------------------------------------------------------
' Host asserts ATT for any command to get device attention.
' First ATT byte contains device id
' Second byte contains cog id and request.
' Next contains addr low, then next contains addr high.
' Last header byte contains length of data.
' Rest of bytes contain data.
'
' BYTE0  DEVID[8]
' BYTE1  COGID[4]CMDREQ[4]
' BYTE2  ADDRLOW[8]
' BYTE3  ADDRHIGH[8]
' BYTE4  LENGTH[8]
' BYTEN  DATA0..N ... for RPC commands, this data is a packet
'
' For header, ACK low means device is ready.
' For transactions, ACK high signals ready for data transfer.
'
' Host Protocol:
' 1 BYTE0 on bus
' 2 Assert ATN low
' 3 Wait for ACK low
' 4 Disassert ATN to high
' 5 Send BYTE1,2,3,4
' 6 Wait for ACK low

' Device Protocol:
' 1 Wait for ATN low
' 2 Check BYTE0 for PropID match
' 3 Assert ACK low
' 4 Prepare to get BYTE1,2,3,4
' 5 Wait for ATN high
' 6 Get BYTE1,2,3,4
' 7 Assert ACK low
' 8 in command assert ACK high

'----------------------------------------------------------------------
nextcmd        ' do next cmd
  andn         dira,    dbusatnstp      ' atn disasserted ... do after wrlong so device detects data stop
  wrlong       zero,    cmdptr          ' let user know command is done
:loop
  rdlong       dst,     dstptr          ' get destination
  and          dst,     #def#PIDMSK     ' only pid bits
  rdbyte       cmd,     cmdptr  wz      ' get command
  tjnz         cmd,     #handleCmd      ' wait for valid command ... if valid
  test         int,     ina     wz      ' if low handle interrupt
  if_nz jmp    #:loop                                                        
  mov          dst,     lastid          ' begin spaghetti poll. set destination for interrupt poll
  mov          cmd,     #def#CMD_POLL   ' set poll command  ... user needs to make sure cmd is clear before next
nextpoll                                ' return point for polling
  test         int,     ina     wz      ' if not low stop handle interrupt
  if_nz jmp    #nextcmd                                                        
handleCmd
  mov          t0,      cmd
  mov          tab,     cmd             ' save cmd to tab for address lookup
  rdlong       adr,     adrptr          ' get target address
  and          t0,      #$f             ' only 3 bits in command
  sub          t0,      #1              ' commands to device are CMD-1 ... 3 bits
  rdlong       t1,      tabptr          ' get command table address
  shl          tab,     #1              ' command addresses are word aligned
  add          t1,      tab             ' add offset for command
  rdlong       len,     lenptr          ' get buffer length
  rdlong       bufp,    bufptr          ' get buffer data pointer

  andn         outa,    leds            ' debug only

  ' 1 BYTE0 on bus
  movs         outa,    dst             ' 1 BYTE0
  waitpeq      ack,     ack             ' ready ?

  ' 2 Assert ATN low
  andn         outa,    atn             ' 1 assert attention bit
  or           dira,    dbusatn         ' assert attention

  ' 3 Wait for ACK low
  waitpne      ack,     ack             ' 3 wait for ack low ... sync window

  ' 4 Disassert ATN to high
  andn         dira,    atn             ' 4 got response, disassert attention

  ' 5 Send BYTE1,2,3,4
  nop
  nop
  nop
  movs         outa,    t0              ' 5 BYTE1 set cogid and modified cmd
  nop
  nop
  mov          outa,    adr             ' 5 BYTE2
  nop
  nop
  shr          outa,    #8              ' 5 BYTE3
  nop
  nop
  movs         outa,    len             ' 5 BYTE4
  nop
  nop
  movs         outa,    #0              ' nothing, just for logic analyzer

  rdword       t0,      t1              ' get command address ... also sync for rd/wr
  jmp          t0                       ' jump to command, returns to nextcmd

'----------------------------------------------------------------------
_idle          ' do nothing command used only at startup 
  jmp          #nextcmd


'----------------------------------------------------------------------
' Fast data transfer with three instructions hub, inst, djnz ... needs a sync point
' Ideas:
' waitpne      stp, stp   ' takes up to 22 cycles, but should catch immediately
'   start data loop       ' can't get it to work
' waitloop until not stp  ' takes 8 cycles ... data loop is 16
'   start data loop       ' works fine

'----------------------------------------------------------------------
' _rdbuf
'
' request must be at least 2 bytes
'
_rdbuf         ' read buffer from target device
  andn         dira,    dbus            ' change bus to input on read
  mov          t1,      #0              ' make sure all t1 input bits are clear
  mov          tp,      bufp            ' set the base address
  add          tp,      len             ' add the length
  or           outa,    leds            ' debug only
:wait
  test         stp,     ina wz          ' wait for stop low
  if_nz jmp    #:wait
  'xor          outa,    leds            ' debug only
  nop
  nop
'{
:read
  movs         t1,      ina wz          ' looking for 0 on all bits and P[0:8]
  wrbyte       t1,      tp              ' write to ptr = address + length
  if_nz djnz   tp,      #:read          ' if t0 <> 0, continue with ptr-1; else stop 
'}
{
:read
  xor          outa,    leds            ' debug only
  rdbyte       t0,      tp              ' dummy debug only
  djnz         len,     #:read          ' if ptr > addr, decrement pointer and repeat
'}
  xor          outa,    leds            ' debug only
  jmp          #nextcmd

'----------------------------------------------------------------------
' _wrbuf
'
' buffer must be at least 2 bytes
'
_wrbuf         ' write buffer to target device
  rdbyte       t0,      t0              ' kill time ... other side has to catch up
  rdbyte       t0,      t0              ' kill time ... other side has to catch up
  rdbyte       t0,      t0              ' kill time ... other side has to catch up
  mov          tp,      bufp            ' set the base address
  add          tp,      len             ' add the length
  rdbyte       t0,      t0              ' sync and kill time ... other side has to catch up
  or           outa,    leds            ' debug only
  andn         outa,    stp             ' set stp low
  or           dira,    stp             ' signal start write ... bit 9 is pulled high externally
  nop                                   ' leave low for stp detect
  xor          outa,    leds            ' debug only
  andn         dira,    stp             ' transfering data ... must be set here

:write         ' write reverse buffer decrementing with P8 controlling read
  rdbyte       outa,    tp              ' get byte from djnz controlled pointer
  cmp          bufp,    tp wz           ' test for end condition
  if_ne djnz   tp,      #:write         ' if ptr > addr, decrement pointer and repeat

  andn         outa,    #$1ff           ' clear bits
  or           dira,    stp             ' signal finished with bits[0:8] == 0
  xor          outa,    leds            ' debug only
  nop                                   ' make sure device gets stop message before releasing it
  rdbyte       t0,      t0              ' make sure device gets stop message before releasing it
  jmp          #nextcmd

'----------------------------------------------------------------------
_runrpc        ' execute a previously arranged remote procedure function call
  jmp          #nextcmd

'----------------------------------------------------------------------
_regrpc        ' register a remote procedure function call
  jmp          #nextcmd


'----------------------------------------------------------------------
_poll          ' poll devices to see who generated ACK interrupt
  waitpeq      ack,     ack             ' ready ?
  mov          t0,      ina             ' get pid
  and          t0,      dbus wz         ' only 8 bits
  if_z jmp     #:decr
  rdlong       head,    hptr            ' get head value
  add          head,    #1              ' add to event queue
  and          head,    #QMSK           ' QMSK = 2**n - 1
  mov          t1,      qptr            ' get qptr base
  add          t1,      head
  wrlong       t0,      t1              ' write back notifier pid
  wrlong       head,    hptr            ' new head value
:decr
  sub          dst,     #1              ' next device
  tjnz         dst,     #nextpoll       ' next device poll
  jmp          #nextcmd                 ' all done. start fresh.


'----------------------------------------------------------------------
dat     ' debug command table
  long         ' align to long
  cmdtable     word _idle       ' 0                     - IDLE command
               word _rdbuf      ' 1 def#CMD_RDBUF       - read buffer from target
               word _wrbuf      ' 2 def#CMD_WRBUF       - write buffer to target
               word _runrpc     ' 3 def#CMD_RUN_RPC     - execute a remote function
               word _regrpc     ' 4 def#CMD_REGISTER    - register a remote function
               word _poll       ' 5 def#CMD_POLL        - poll devices for event
               word _idle       ' 6 def#CMD_NOP         - used only at startup
               word _idle       ' 7 def#CMD_NOP         - used only at startup
               word _idle       ' 8 def#CMD_NOP         - used only at startup
              
{{
 TERMS OF USE: MIT License                                                           

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
}}